import { Metadata, FormatterParserProvider, Http, Validation, RestDatasource } from '@zeedhi/core';
import { Modal, ModalService, TreeGrid, TreeGridEditable, Form, LoadingService, AlertService, DialogService, SelectTree } from '@zeedhi/common';
import { CustomAlertService } from '@/Services/CustomAlertService';
import { Utilities } from '@/Services/Utilities';

FormatterParserProvider.registerFormatter('column_SelectMultipleFormatter', (params: any) => {
    const formatterFn = FormatterParserProvider.getFormatter('column_ZdSelect');
    const oldValue = params.value;
    if (params.componentProps.datasource.data) {
        params.column.loading = false;
    }
    if (!oldValue?.length) {
        return "";
    }
    const formatedValue = oldValue.map((value: any) => formatterFn({...params, value: [value]})).map((val: any) => (val instanceof Array ? val[0] : val));
    /*  if (formatedValue.every((v: any, i: any) => v === oldValue[i])) {
        return "";
    } */
    return formatedValue.join(', ');
});

export class MenuConfigurationController {
    //  Metodos de dialog
    public dialogEvents: any = {
        click: async ({ component }: any) => {
            if (component.name === "sim") {
                await this.deleteRow();
            }
            DialogService.hide();
        }
    };
    public showDeleteDialog({ event }: any) {
        event.preventDefault();
        DialogService.show({
          name: 'delete-dialog',
          type: 'warning',
          title: ' ',
          text: 'Deseja remover o(s) registro(s) selecionado(s)?',
          buttons: [
            { name: 'não', component: 'ZdButton', label: 'NÃO', events: this.dialogEvents, outline: true },
            { name: 'sim', component: 'ZdButton', label: 'SIM', events: this.dialogEvents },
          ],
        });
    }
    //  Metodos de Modal
    private detailModal!: Modal;
    private filterModal!: Modal;
    public async createDetailModal() {
        this.detailModal = await ModalService.createFromJson('menuConfiguration/DetailModal/DetailModal', true);
    }
    public async createFilterModal() {
        this.filterModal = await ModalService.createFromJson('menuConfiguration/FilterModal/FilterModal', true);
    }
    //  FilterForm
    get isDatasourceWithoutFilter() {
        const grid = Metadata.getInstance<TreeGrid>(this.gridName);
        const datasource = grid.datasource;
        const filters = {...datasource.filter};
        filters.type = undefined; // Ignora type: "M"
        return !Object.values(filters).reduce((el, filtro) => filtro || el, false);
    }
    public filterOnMounted({ component }: any) {
        const filter = Metadata.getInstance<TreeGrid>(this.gridName).datasource.filter;
        const form = component;
        const formValue = form.value;
        formValue.name = filter.name ? filter.name.slice(1, -1) : "";
        formValue.order = filter._order;
        form.getChildInstance('accessProfiles').value = filter.accessProfiles;
    }
    public async openFilterModal() {
        this.filterModal?.destroy();
        await this.createFilterModal();
        this.filterModal.show();
    }
    public async applyFilter() {
        const grid = Metadata.getInstance<TreeGrid>(this.gridName);
        const datasource = grid.datasource;
        const form = this.filterModal.getChildInstance<Form>("menuGrid-filter-form");
        const values = form.value;
        const defaultFilter = {"type": "M"};
        const filters = {
            "_order": values.order ? Number(values.order) : undefined, //  "order" é uma chave reservada do Zeedhi, sendo assim decidi tratar este parametro isoladamente e criar um caso exclusivo no backend
            "name": values.name ? "%" + values.name + "%" : undefined,
            "accessProfiles": values.accessProfiles.length > 0 ? values.accessProfiles : undefined
        };
        // Se houver algum filtro com valor
        if (Object.values(filters).reduce((el, filtro) => filtro || el)) {
            const result = await datasource.setFilter({...filters, ...defaultFilter});
        } else {
            const res = await datasource.setFilter(defaultFilter);
            this.globalGridData = res;
        }
        this.closeFilter();
    }
    public closeDetailModal() {
        this.detailModal?.hide();
    }
    public closeFilter() {
        this.filterModal?.hide();
    }
    //  Util
    public globalGridData: any;
    public async loadGlobalGridData(loadedData: any) {
        if (loadedData && loadedData.length > 0) {
            this.globalGridData = loadedData;
            return loadedData;
        }
        const grid = Metadata.getInstance<TreeGrid>(this.gridName);
        const datasource = new RestDatasource(grid.datasource);
        datasource.setFilter({"type": "M"});
        const res = await datasource.get();
        this.globalGridData = res;
        return res;
    }
    public getTreeGridDataDict() {
        const data = this.globalGridData || Metadata.getInstance<TreeGrid>(this.gridName).datasource.data;
        const dataDict = this.generateDataDict(data);
        this.updateInfo(Object.values(dataDict).filter(({parent}: any) => !parent)); // Apenas os elementos sem pai (método cascata)
        return dataDict;
    }
    public generateDataDict(data: any, idToIgnore?: any) {
        const dataDict: any = {};
        //  Creates the .children property
        data.forEach((item: any) => {
            let obj: any = dataDict[item._id] || {};
            obj = {...item, ...obj};
            dataDict[obj._id] = obj;
            if (obj.parent) {
                const parent = dataDict[obj.parent] || {};
                if (!parent.children) {
                    parent.children = [obj];
                } else {
                    parent.children.push(obj);
                }
                dataDict[obj.parent] = parent;
            }
        });
        //  If needed to ignore certain id
        if (idToIgnore) {
            const obj = dataDict[idToIgnore];
            if (obj.parent) {
                const parent = dataDict[obj.parent];
                parent.children = parent.children.filter(({id}: any) => id !== idToIgnore);
                if (parent.children.length === 0) {
                    parent.children = undefined;
                }
            }
            delete dataDict[idToIgnore];
        }
        return dataDict;
    }
    public treeGridToTreeSelect(idToIgnore: any) {
        const data = this.globalGridData || Metadata.getInstance<TreeGrid>(this.gridName).datasource.data;
        const extraData = data.map((item: any) => ({...item, id: item._id, label: item.name, parent: item.parent, children: item.children}));
        const dataDict = this.generateDataDict(extraData, idToIgnore);
        return Object.values(dataDict).filter(({parent}: any) => !parent);
    }
    public updateInfo(childrenDatas: any, level: any = 0) {
        return childrenDatas.reduce((arr: any, child: any) => {
            child._level = level + 1;
            child._childCount = child.children?.length || 0;
            arr.push(child);
            if (child._childCount > 0) {
                const res = this.updateInfo(child.children, child._level);
                arr.push(...res);
            }
            return arr;
        }, []);
    }
    public changeParent(children: any, parent?: any) {
        const parentId = parent?._id;
        children.forEach((child: any) => {
            child.parent = parentId;
            if (parent) {
                if (parent.children) {
                    parent.children.push(child);
                } else {
                    parent.children = [child];
                }
                parent._childCount = parent.children.length;
            }
        });
        return this.updateInfo(children, parent?._level);
    }
    //  DetailForm
    protected modalData: any;
    public get currentRowId() {
        return this.modalData._id;
    }
    public get modalTitle() {
        return (this.currentRowId ? "Editar" : "Adicionar") + " Menu";
    }
    public async formOnMounted({ component }: any) {
        const form = component;
        const formValue = form.value;
        const data = this.modalData;
        formValue.name = data.name;
        formValue.order = data.order;
        formValue.created_by = this.currentRowId ? data.created_by_name : this.user.username;
        // parent Field
        const parent = form.getChildInstance('parent');
        parent.setValue(data.parent);
        parent.disabledItems = [data._id];
        // accessProfiles Field
        const accessProfiles = form.getChildInstance('accessProfiles');
        if (data.accessProfiles?.length > 0) {
            //  O SelectMultiple efetua uma requisição por item, o accessProfiles demonstra apresentar vários itens, desta forma todos os itens são retornados em uma unica requisição
            const apDatasource = new RestDatasource(accessProfiles.datasource);
            apDatasource.searchIn = [apDatasource.uniqueKey];
            apDatasource.search = data.accessProfiles;
            const result = await apDatasource.get();
            accessProfiles.datasource.data = result;
        }
        accessProfiles.setValue(data.accessProfiles);
    }
    public async openDetailModal({ event, component, row}: any) {
        //  Checa se o evento foi de fato expandir a linha ao invés de click na mesma
        if (event.target.type === "button" || component.editing) {
            return;
        }
        this.detailModal?.destroy();
        await this.createDetailModal();
        this.modalData = { type: "M", ...row};
        if (row) {
            Object.keys(row).filter((key) => key && key.startsWith("tree__")).forEach((key) => delete this.modalData[key]);
        }
        this.detailModal.show();
    }
    public generateSaveData(values: any, preValues: any = {}) {
        const dataDict = this.getTreeGridDataDict();
        const details = values._id ? dataDict[values._id] : undefined;
        const content = {
            ...preValues,
            type: "M",
            name: values.name,
            accessProfiles: values.accessProfiles?.length > 0 ? values.accessProfiles : null,
            order: values.order ? Number(values.order) : null,
            parent: values.parent,
            children: details?.children,
            _childCount: details?._childCount || 0,
            _level: (values.parent ? dataDict[values.parent]._level : 0) + 1,
            _id: values._id,
            "changed": true
        };
        return values._id ? content : {
            __createdLocal: true,
            __is_new: true,
            created_by: this.user.userId,
            created_by_name: this.user.username,
            ...content
        };
    }
    public updateParents(row: any, previousId: string = "", newId: string = "") {
        const content = [];
        const dataDict = this.getTreeGridDataDict();
        if (previousId !== newId) {
            if (previousId) {
                //  Atualiza o childCount do pai que não é mais pai
                const pastParentRow = dataDict[previousId];
                pastParentRow._childCount -= 1;
                content.push(pastParentRow);
            }
            if (newId) {
                const newParentRow = dataDict[newId];
                content.push(...this.changeParent([row], newParentRow), newParentRow);
            } else {
                content.push(...this.changeParent([row]));
            }
        }
        return content;
    }
    public cleanUpDataset(dataset: any) {
        const originalValues = this.getTreeGridDataDict();
        const cleaned: any[] = Array.from(new Set(dataset)).map((data: any) => { delete data.children; return data; });
        return cleaned.filter((el: any) => {
            const oldValues = originalValues[el._id];
            return !el._id || !Object.keys(el).reduce((isEqual: boolean, key: any) => isEqual && el[key] === oldValues[key], true);
        });
    }
    public async saveDetailModal() {
        const form = this.detailModal.getChildInstance<Form>("detailForm");
        const values = form.value;
        if (!form.validate()) {
            return;
        }
        const currentRow = this.generateSaveData({
            ...values,
            // @ts-ignore
            parent: form.getChildInstance<SelectTree>('parent').value,
            _id: this.currentRowId
        }, this.modalData);
        let content = [...this.updateParents(currentRow, this.modalData.parent, currentRow.parent), currentRow];
        content = this.cleanUpDataset(content);
        try {
            const res = await Http.post("view/save", {
                "requestType": "DataSet",
                "disableLoader": false,
                "dataset": content
            });
            CustomAlertService.saveSuccessAlert();
        } catch (err) {
            CustomAlertService.errorAlert("Ocorreu um erro.");
            console.error(err);
        }
        this.reload();
        this.detailModal.hide();
    }
    //  Grid Methods
    public gridName = "menuGrid";
    get deleteButtonDisabled(): boolean {
        const grid = Metadata.getInstance<TreeGrid>(this.gridName);
        return grid.selectedRows.length <= 0;
    }
    protected user: any;
    public async gridOnMounted() {
        const grid = Metadata.getInstance<TreeGrid>(this.gridName);
        await this.reload();
        Utilities.getLoggedUser().then((res) => {
            const login = res.data;
            this.user = {
                userId: login.userId,
                username: login.username
            };
        });
    }
    public rowsToUpdateAfterDeletion(deletedRows: any) {
        const deletedIds = deletedRows.map((el: any) => el._id);
        const dataDict = this.getTreeGridDataDict();
        const rowsToUpdate = deletedRows.reduce((changedRows: any, el: any) => {
            const children = dataDict[el._id]?.children?.filter(({_id}: any) => !deletedIds.includes(_id)) || [];
            let parent = el.parent;
            let updateParent = false;
            while (parent && deletedIds.includes(parent)) {
                parent = dataDict[parent]?.parent || undefined;
            }
            const parentEl = dataDict[parent] || undefined;
            if (parent && el.parent === parent) {
                 //  Caso o pai que não será deletado for o pai direto do elemento a ser deletado...
                const i = parentEl.children?.map(({_id}: any) => _id).indexOf(el._id);
                if (i > -1) {
                    parentEl.children.splice(i, 1);
                    parentEl._childCount -= 1;
                    updateParent = true;
                }
            }
            changedRows.push(...this.changeParent(children, parentEl));
            if (updateParent || (parent && children.length > 0)) {
                changedRows.push(parentEl);
            }
            return changedRows;
        }, []).filter((el: any) => el);
        return this.cleanUpDataset(rowsToUpdate);
    }
    public async deleteRow() {
        const grid = Metadata.getInstance<TreeGrid>(this.gridName);
        const selectedRows = grid.selectedRows;
        try {
            await Http.post("/view/remove", {
                "requestType": "DataSet",
                "dataset": selectedRows
            });
            CustomAlertService.deleteSuccessAlert();
            grid.selectedRows = [];
        } catch (err) {
            console.error(err);
            CustomAlertService.errorAlert("Ocorreu um erro.");
        }
        const rowsToUpdate = this.rowsToUpdateAfterDeletion(selectedRows);
        if (rowsToUpdate.length > 0) {
            try {
                await Http.post("view/save", {
                    "requestType": "DataSet",
                    "disableLoader": false,
                    "dataset": rowsToUpdate
                });
            } catch (err) {
                console.error(err);
                CustomAlertService.errorAlert("Ocorreu um erro.");
            }
        }
        this.reload();
    }
    public async reload() {
        const grid = Metadata.getInstance<TreeGrid>(this.gridName);
        const res = await grid.datasource.get();
        if (this.isDatasourceWithoutFilter) {
            await this.loadGlobalGridData(res);
        }
    }
    public get isEditing() {
        return Metadata.getInstance<TreeGridEditable>(this.gridName).editing;
    }
    public get isNotEditing() {
        return !this.isEditing;
    }
    public cancelInlineEdit() {
        Metadata.getInstance<TreeGridEditable>(this.gridName).cancelEditedRows();
    }
    public async saveInlineEdit() {
        const grid = Metadata.getInstance<TreeGridEditable>(this.gridName);
        const rows = grid.getEditedRows();
        rows.forEach((row: any) => { row.changed = true; });
        try {
            const res = await Http.post("view/save", {
                "requestType": "DataSet",
                "disableLoader": false,
                "dataset": rows
            });
            CustomAlertService.saveSuccessAlert();
        } catch (err) {
            CustomAlertService.errorAlert("Ocorreu um erro.");
            console.error(err);
        }
        grid.saveEditedRows();
    }
}
